#!/hint/bash
#
# SPDX-License-Identifier: GPL-3.0-or-later

[[ -z ${DEVTOOLS_INCLUDE_BUILD_SH:-} ]] || return 0
DEVTOOLS_INCLUDE_BUILD_SH=1

_DEVTOOLS_LIBRARY_DIR=${_DEVTOOLS_LIBRARY_DIR:-/usr/local/share/devtools-git-poc}
# shellcheck source=src/lib/common.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/common.sh
# shellcheck source=src/lib/util/pacman.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/util/pacman.sh
# shellcheck source=src/lib/valid-repos.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-repos.sh
# shellcheck source=src/lib/valid-tags.sh
source "${_DEVTOOLS_LIBRARY_DIR}"/lib/valid-tags.sh

source /usr/share/makepkg/util/config.sh
source /usr/share/makepkg/util/message.sh

set -e


pkgctl_build_usage() {
	local -r COMMAND=${_DEVTOOLS_COMMAND:-${BASH_SOURCE[0]##*/}}
    cat <<- _EOF_
		Usage: ${COMMAND} [OPTIONS] [PATH]

		Build packages inside a clean chroot

		When a new pkgver is set using the appropriate PKGBUILD options the
		checksums are automatically updated.

		TODO

		BUILD OPTIONS
		    --arch ARCH          Specify architectures to build for (disables auto-detection)
		    --repo REPO          Specify a target repository (disables auto-detection)
		    -s, --staging        Build against the staging counterpart of the auto-detected repo
		    -t, --testing        Build against the testing counterpart of the auto-detected repo
		    -o, --offload        Build on a remote server and transfer artifacts afterwards
		    -c, --clean          Recreate the chroot before building
		    -I, --install FILE   Install a package into the working copy of the chroot

		PKGBUILD OPTIONS
		    --pkgver=PKGVER      Set the pkgver variable and reset pkgrel to 1
		    --pkgrel=PKGREL      Set the pkgrel variable to a given value
		    --rebuild            Increment the pkgrel variable by 1
		    -e, --edit           Edit the PKGBUILD before building

		RELEASE OPTIONS
		    -r, --release        Automatically commit, tag and release after building
		    -m, --message MSG    Use the given <msg> as the commit message
		    -u, --db-update      Automatically update the pacman database after uploading

		OPTIONS
		    -h, --help           Show this help text

		EXAMPLES
		    $ ${COMMAND}
		    $ ${COMMAND} --rebuild --staging
		    $ ${COMMAND} --pkgver 1.42 --release --db-update
_EOF_
}

pkgctl_build_check_option_group_repo() {
	local option=$1
	local repo=$2
	local testing=$3
	local staging=$4
	if [[ -n "${repo}" ]] || (( testing )) || (( staging )); then
		die "The argument '%s' cannot be used with one or more of the other specified arguments" "${option}"
		exit 1
	fi
	return 0
}

pkgctl_build_check_option_group_ver() {
	local option=$1
	local pkgver=$2
	local pkgrel=$3
	local rebuild=$4
	if [[ -n "${pkgver}" ]] || [[ -n "${pkgrel}" ]] || (( rebuild )); then
		die "The argument '%s' cannot be used with one or more of the other specified arguments" "${option}"
		exit 1
	fi
	return 0
}

# TODO: import pgp keys
pkgctl_build() {
	if (( $# < 1 )) && [[ ! -f PKGBUILD ]]; then
		pkgctl_build_usage
		exit 1
	fi

	local UPDPKGSUMS=0
	local EDIT=0
	local REBUILD=0
	local OFFLOAD=0
	local STAGING=0
	local TESTING=0
	local RELEASE=0

	local PKGDIR="${PWD}"
	local REPO=
	local PKGVER=
	local PKGREL=
	local MESSAGE=

	local BUILD_ARCH=()
	local BUILD_OPTIONS=()
	local MAKECHROOT_OPTIONS=()
	local RELEASE_OPTIONS=()

	local PTS
	PTS="$(tty | sed 's|/dev/pts/||')"
	local WORKER="${USER}-${PTS}"

	while (( $# )); do
		case $1 in
			-h|--help)
				pkgctl_build_usage
				exit 0
				;;
			--repo)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				REPO="${2}"
				shift 2
				;;
			--arch)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				if [[ ${2} == all ]]; then
					BUILD_ARCH=("${_arch[@]}")
				elif [[ ${2} == any ]]; then
					BUILD_ARCH=("${_arch[0]}")
				elif ! in_array "${2}" "${BUILD_ARCH[@]}"; then
					if ! in_array "${2}" "${_arch[@]}"; then
						die 'invalid architecture: %s' "${2}"
					fi
					BUILD_ARCH+=("${2}")
				fi
				shift 2
				;;
			--pkgver=*)
				pkgctl_build_check_option_group_ver '--pkgver' "${PKGVER}" "${PKGREL}" "${REBUILD}"
				PKGVER="${1#*=}"
				PKGREL=1
				UPDPKGSUMS=1
				shift
				;;
			--pkgrel=*)
				pkgctl_build_check_option_group_ver '--pkgrel' "${PKGVER}" "${PKGREL}" "${REBUILD}"
				PKGREL="${1#*=}"
				shift
				;;
			--rebuild)
				pkgctl_build_check_option_group_ver '--rebuild' "${PKGVER}" "${PKGREL}" "${REBUILD}"
				REBUILD=1
				shift
				;;
			-e|--edit)
				EDIT=1
				shift
				;;
			-o|--offload)
				OFFLOAD=1
				shift
				;;
			-s|--staging)
				pkgctl_build_check_option_group_repo '--staging' "${REPO}" "${TESTING}" "${STAGING}"
				STAGING=1
				shift
				;;
			-t|--testing)
				pkgctl_build_check_option_group_repo '--testing' "${REPO}" "${TESTING}" "${STAGING}"
				TESTING=1
				shift
				;;
			-c|--clean)
				BUILD_OPTIONS+=("-c")
				shift
				;;
			-I|--install)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				MAKECHROOT_OPTIONS+=("-I" "$2")
				warning 'installing packages into the chroot may break reproducible builds, use with caution!'
				shift 2
				;;
			-r|--release)
				# shellcheck source=src/lib/release.sh
				source "${_DEVTOOLS_LIBRARY_DIR}"/lib/release.sh
				RELEASE=1
				shift
				;;
			-m|--message)
				(( $# <= 1 )) && die "missing argument for %s" "$1"
				MESSAGE=$2
				RELEASE_OPTIONS+=("--message" "${MESSAGE}")
				shift 2
				;;
			-u|--db-update)
				RELEASE_OPTIONS+=("--db-update")
				shift
				;;
			--)
				shift
				break
				;;
			-*)
				die "invalid argument: %s" "$1"
				;;
			*)
				[[ ! -d $1 ]] || [[ ! -f $1/PKGBUILD ]] && die "Not a package directory: %s" "$1"
				PKGDIR="$1"
				shift
				;;
		esac
	done

	# By default build in current working dir
	pushd "${PKGDIR}" >/dev/null

	# make PKGBUILD vars available
	if [[ ! -f PKGBUILD ]]; then
		die 'PKGBUILD not found'
	fi
	# shellcheck source=contrib/makepkg/PKGBUILD.proto
	. ./PKGBUILD

	# increment pkgrel on rebuild
	if (( REBUILD )); then
		PKGREL=$((pkgrel+1))
	fi

	# update pkgver
	if [[ -n ${PKGVER} ]]; then
		if [[ $(type -t pkgver) == function ]]; then
			# TODO: check if die or warn, if we provide _commit _gitcommit setter maybe?
			warning 'setting pkgver variable has no effect if the PKGBUILD has a pkgver() function'
		fi
		msg "Bumping pkgver to ${PKGVER}"
		grep -Eqm1 "^pkgver=${pkgver}$" PKGBUILD || die "Non-standard pkgver declaration"
		sed -E "s|^(pkgver=)${pkgver}$|\1${PKGVER}|g" -i PKGBUILD
	fi

	# update pkgrel
	if [[ -n ${PKGREL} ]]; then
		msg "Bumping pkgrel to ${PKGREL}"
		grep -Eqm1 "^pkgrel=${pkgrel}$" PKGBUILD || die "Non-standard pkgrel declaration"
		sed -E "s|^(pkgrel=)${pkgrel}$|\1${PKGREL}|g" -i PKGBUILD
	fi

	# edit PKGBUILD
	if (( EDIT )); then
		stat_busy 'Editing PKGBUILD'
		if [[ -n $VISUAL ]]; then
			$VISUAL PKGBUILD || die
		elif [[ -n $EDITOR ]]; then
			$EDITOR PKGBUILD || die
		elif command -v vi &>/dev/null; then
			vi PKGBUILD || die
		else
			die "Need one of \$VISUAL or \$EDITOR to edit the PKGBUILD"
		fi
		stat_done
	fi

	# update checksums
	if (( UPDPKGSUMS )) && (( ${#source[@]} >= 1 )); then
		updpkgsums
	fi

	# auto-detection of build target
	if [[ -z ${REPO} ]]; then
		update_pacman_repo_cache
		if ! REPO=$(get_pacman_repo_from_pkgbuild PKGBUILD); then
			die 'Failed to get pacman repo'
		fi
		if [[ -z "${REPO}" ]]; then
			die 'Unknown repo, please specify --repo for new packages'
		fi
	elif ! in_array "${REPO}" "${_repos[@]}"; then
		die "Invalid repository target: %s" "${REPO}"
	fi

	# TODO: only during PoC
	if [[ $REPO == community ]]; then
		REPO=extra
	fi

	# special cases to resolve final build target
	if (( TESTING )); then
		REPO="${REPO}-testing"
	elif (( STAGING )); then
		REPO="${REPO}-staging"
	elif [[ $REPO == core ]]; then
		REPO="${REPO}-testing"
	fi

	# auto-detection of build architecture
	if [[ $REPO = multilib* ]]; then
		BUILD_ARCH=("")
	elif (( ${#BUILD_ARCH[@]} == 0 )); then
		if in_array any "${arch[@]}"; then
			BUILD_ARCH=("${_arch[0]}")
		else
			BUILD_ARCH+=("${arch[@]}")
		fi
	fi

	# execute build
	pkgbase=$(basename "${PKGDIR}")
	for arch in "${BUILD_ARCH[@]}"; do
		[[ $arch == any ]] && continue

		if [[ -n $arch ]]; then
			msg "Building ${pkgbase} for [${REPO}] (${arch})"
			BUILDTOOL="${REPO}-${arch}-build"
		else
			msg "Building ${pkgbase} for [${REPO}]"
			BUILDTOOL="${REPO}-build"
		fi

		if (( OFFLOAD )); then
			offload-build --repo "${REPO}" -- "${BUILD_OPTIONS[@]}" -- "${MAKECHROOT_OPTIONS[@]}" -l "${WORKER}"
		else
			"${BUILDTOOL}" "${BUILD_OPTIONS[@]}" -- "${MAKECHROOT_OPTIONS[@]}" -l "${WORKER}"
		fi
	done

	# release the build
	if (( RELEASE )); then
		pkgctl_release --repo "${REPO}" "${RELEASE_OPTIONS[@]}"
	fi
}
